@@ -27,23 +27,22 @@
  * Paste paste buffer if present.
  */
 
-enum cmd_retval	 cmd_paste_buffer_exec(struct cmd *, struct cmd_ctx *);
+enum cmd_retval	 cmd_paste_buffer_exec(struct cmd *, struct cmd_q *);
 
 void	cmd_paste_buffer_filter(struct window_pane *,
-	    const char *, size_t, const char *, int bracket);
+	    const char *, size_t, const char *, int);
 
 const struct cmd_entry cmd_paste_buffer_entry = {
 	"paste-buffer", "pasteb",
 	"db:prs:t:", 0, 0,
-	"[-dpr] [-s separator] [-b buffer-index] [-t target-pane]",
+	"[-dpr] [-s separator] [-b buffer-index] " CMD_TARGET_PANE_USAGE,
 	0,
 	NULL,
-	NULL,
 	cmd_paste_buffer_exec
 };
 
 enum cmd_retval
-cmd_paste_buffer_exec(struct cmd *self, struct cmd_ctx *ctx)
+cmd_paste_buffer_exec(struct cmd *self, struct cmd_q *cmdq)
 {
 	struct args		*args = self->args;
 	struct window_pane	*wp;
@@ -54,7 +53,7 @@ cmd_paste_buffer_exec(struct cmd *self, struct cmd_ctx *ctx)
 	int			 buffer;
 	int			 pflag;
 
-	if (cmd_find_pane(ctx, args_get(args, 't'), &s, &wp) == NULL)
+	if (cmd_find_pane(cmdq, args_get(args, 't'), &s, &wp) == NULL)
 		return (CMD_RETURN_ERROR);
 
 	if (!args_has(args, 'b'))
@@ -62,7 +61,7 @@ cmd_paste_buffer_exec(struct cmd *self, struct cmd_ctx *ctx)
 	else {
 		buffer = args_strtonum(args, 'b', 0, INT_MAX, &cause);
 		if (cause != NULL) {
-			ctx->error(ctx, "buffer %s", cause);
+			cmdq_error(cmdq, "buffer %s", cause);
 			free(cause);
 			return (CMD_RETURN_ERROR);
 		}
@@ -73,7 +72,7 @@ cmd_paste_buffer_exec(struct cmd *self, struct cmd_ctx *ctx)
 	else {
 		pb = paste_get_index(&global_buffers, buffer);
 		if (pb == NULL) {
-			ctx->error(ctx, "no buffer %d", buffer);
+			cmdq_error(cmdq, "no buffer %d", buffer);
 			return (CMD_RETURN_ERROR);
 		}
 	}
@@ -86,9 +85,8 @@ cmd_paste_buffer_exec(struct cmd *self, struct cmd_ctx *ctx)
 			else
 				sepstr = "\r";
 		}
-		pflag = args_has(args, 'p') &&
-		    (wp->screen->mode & MODE_BRACKETPASTE);
-		cmd_paste_buffer_filter(wp, pb->data, pb->size, sepstr, pflag);
+		pflag = (wp->screen->mode & MODE_BRACKETPASTE);
+		paste_send_pane(pb, wp, sepstr, args_has(args, 'p') && pflag);
 	}
 
 	/* Delete the buffer if -d. */
@@ -101,30 +99,3 @@ cmd_paste_buffer_exec(struct cmd *self, struct cmd_ctx *ctx)
 
 	return (CMD_RETURN_NORMAL);
 }
-
-/* Add bytes to a buffer and filter '\n' according to separator. */
-void
-cmd_paste_buffer_filter(struct window_pane *wp,
-    const char *data, size_t size, const char *sep, int bracket)
-{
-	const char	*end = data + size;
-	const char	*lf;
-	size_t		 seplen;
-
-	if (bracket)
-		bufferevent_write(wp->event, "\033[200~", 6);
-
-	seplen = strlen(sep);
-	while ((lf = memchr(data, '\n', end - data)) != NULL) {
-		if (lf != data)
-			bufferevent_write(wp->event, data, lf - data);
-		bufferevent_write(wp->event, sep, seplen);
-		data = lf + 1;
-	}
-
-	if (end != data)
-		bufferevent_write(wp->event, data, end - data);
-
-	if (bracket)
-		bufferevent_write(wp->event, "\033[201~", 6);
-}
