@@ -35,7 +35,6 @@ const struct cmd_entry cmd_swap_pane_entry = {
 	"[-dDU] " CMD_SRCDST_PANE_USAGE,
 	0,
 	cmd_swap_pane_key_binding,
-	NULL,
 	cmd_swap_pane_exec
 };
 
@@ -75,8 +74,12 @@ cmd_swap_pane_exec(struct cmd *self, struct cmd_q *cmdq)
 			src_wp = TAILQ_PREV(dst_wp, window_panes, entry);
 			if (src_wp == NULL)
 				src_wp = TAILQ_LAST(&dst_w->panes, window_panes);
-		} else
-			return (CMD_RETURN_NORMAL);
+		} else {
+			src_wl = cmd_find_pane(cmdq, NULL, NULL, &src_wp);
+			if (src_wl == NULL)
+				return (CMD_RETURN_ERROR);
+			src_w = src_wl->window;
+		}
 	} else {
 		src_wl = cmd_find_pane(cmdq, args_get(args, 's'), NULL, &src_wp);
 		if (src_wl == NULL)
