@@ -28,7 +28,7 @@
  */
 
 void		 cmd_switch_client_key_binding(struct cmd *, int);
-enum cmd_retval	 cmd_switch_client_exec(struct cmd *, struct cmd_ctx *);
+enum cmd_retval	 cmd_switch_client_exec(struct cmd *, struct cmd_q *);
 
 const struct cmd_entry cmd_switch_client_entry = {
 	"switch-client", "switchc",
@@ -36,7 +36,6 @@ const struct cmd_entry cmd_switch_client_entry = {
 	"[-lnpr] [-c target-client] [-t target-session]",
 	CMD_READONLY,
 	cmd_switch_client_key_binding,
-	NULL,
 	cmd_switch_client_exec
 };
 
@@ -58,47 +57,74 @@ cmd_switch_client_key_binding(struct cmd *self, int key)
 }
 
 enum cmd_retval
-cmd_switch_client_exec(struct cmd *self, struct cmd_ctx *ctx)
+cmd_switch_client_exec(struct cmd *self, struct cmd_q *cmdq)
 {
-	struct args	*args = self->args;
-	struct client	*c;
-	struct session	*s;
+	struct args		*args = self->args;
+	struct client		*c;
+	struct session		*s = NULL;
+	struct winlink		*wl = NULL;
+	struct window 		*w = NULL;
+	struct window_pane	*wp = NULL;
+	const char		*tflag;
 
-	if ((c = cmd_find_client(ctx, args_get(args, 'c'))) == NULL)
+	if ((c = cmd_find_client(cmdq, args_get(args, 'c'), 0)) == NULL)
 		return (CMD_RETURN_ERROR);
 
 	if (args_has(args, 'r')) {
 		if (c->flags & CLIENT_READONLY) {
 			c->flags &= ~CLIENT_READONLY;
-			ctx->info(ctx, "made client writable");
+			cmdq_info(cmdq, "made client writable");
 		} else {
 			c->flags |= CLIENT_READONLY;
-			ctx->info(ctx, "made client read-only");
+			cmdq_info(cmdq, "made client read-only");
 		}
 	}
 
-	s = NULL;
+	tflag = args_get(args, 't');
 	if (args_has(args, 'n')) {
 		if ((s = session_next_session(c->session)) == NULL) {
-			ctx->error(ctx, "can't find next session");
+			cmdq_error(cmdq, "can't find next session");
 			return (CMD_RETURN_ERROR);
 		}
 	} else if (args_has(args, 'p')) {
 		if ((s = session_previous_session(c->session)) == NULL) {
-			ctx->error(ctx, "can't find previous session");
+			cmdq_error(cmdq, "can't find previous session");
 			return (CMD_RETURN_ERROR);
 		}
 	} else if (args_has(args, 'l')) {
 		if (c->last_session != NULL && session_alive(c->last_session))
 			s = c->last_session;
 		if (s == NULL) {
-			ctx->error(ctx, "can't find last session");
+			cmdq_error(cmdq, "can't find last session");
 			return (CMD_RETURN_ERROR);
 		}
-	} else
-		s = cmd_find_session(ctx, args_get(args, 't'), 0);
-	if (s == NULL)
-		return (CMD_RETURN_ERROR);
+	} else {
+		if (tflag == NULL) {
+			if ((s = cmd_find_session(cmdq, tflag, 1)) == NULL)
+				return (CMD_RETURN_ERROR);
+		} else if (tflag[strcspn(tflag, ":.")] != '\0') {
+			if ((wl = cmd_find_pane(cmdq, tflag, &s, &wp)) == NULL)
+				return (CMD_RETURN_ERROR);
+		} else {
+			if ((s = cmd_find_session(cmdq, tflag, 1)) == NULL)
+				return (CMD_RETURN_ERROR);
+			w = cmd_lookup_windowid(tflag);
+			if (w == NULL &&
+			    (wp = cmd_lookup_paneid(tflag)) != NULL)
+				w = wp->window;
+			if (w != NULL)
+				wl = winlink_find_by_window(&s->windows, w);
+		}
+
+		if (cmdq->client == NULL)
+			return (CMD_RETURN_NORMAL);
+
+		if (wl != NULL) {
+			if (wp != NULL)
+				window_set_active_pane(wp->window, wp);
+			session_set_current(s, wl);
+		}
+	}
 
 	if (c->session != NULL)
 		c->last_session = c->session;
