@@ -37,7 +37,6 @@
 
 /* Default grid cell data. */
 const struct grid_cell grid_default_cell = { 0, 0, 8, 8, (1 << 4) | 1, " " };
-const struct grid_cell grid_marker_cell = { 0, 0, 8, 8, (1 << 4) | 1, "_" };
 
 #define grid_put_cell(gd, px, py, gc) do {			\
 	memcpy(&gd->linedata[py].celldata[px], 			\
@@ -124,7 +123,7 @@ grid_compare(struct grid *ga, struct grid *gb)
 	struct grid_cell	*gca, *gcb;
 	u_int			 xx, yy;
 
-	if (ga->sx != gb->sx || ga->sy != ga->sy)
+	if (ga->sx != gb->sx || ga->sy != gb->sy)
 		return (1);
 
 	for (yy = 0; yy < ga->sy; yy++) {
@@ -268,8 +267,7 @@ grid_get_cell(struct grid *gd, u_int px, u_int py)
 
 /* Set cell at relative position. */
 void
-grid_set_cell(
-    struct grid *gd, u_int px, u_int py, const struct grid_cell *gc)
+grid_set_cell(struct grid *gd, u_int px, u_int py, const struct grid_cell *gc)
 {
 	if (grid_check_y(gd, py) != 0)
 		return;
@@ -592,6 +590,7 @@ grid_string_cells(struct grid *gd, u_int px, u_int py, u_int nx,
 	char			*buf, code[128];
 	size_t			 len, off, size, codelen;
 	u_int			 xx;
+	const struct grid_line	*gl;
 
 	GRID_DEBUG(gd, "px=%u, py=%u, nx=%u", px, py, nx);
 
@@ -604,8 +603,11 @@ grid_string_cells(struct grid *gd, u_int px, u_int py, u_int nx,
 	buf = xmalloc(len);
 	off = 0;
 
+	gl = grid_peek_line(gd, py);
 	for (xx = px; xx < px + nx; xx++) {
-		gc = grid_peek_cell(gd, xx, py);
+		if (gl == NULL || xx >= gl->cellsize)
+			break;
+		gc = &gl->celldata[xx];
 		if (gc->flags & GRID_FLAG_PADDING)
 			continue;
 		grid_cell_get(gc, &ud);
@@ -641,7 +643,7 @@ grid_string_cells(struct grid *gd, u_int px, u_int py, u_int nx,
         if (trim) {
 		while (off > 0 && buf[off - 1] == ' ')
 			off--;
-        }
+	}
 	buf[off] = '\0';
 
 	return (buf);
@@ -653,8 +655,8 @@ grid_string_cells(struct grid *gd, u_int px, u_int py, u_int nx,
  * available.
  */
 void
-grid_duplicate_lines(
-    struct grid *dst, u_int dy, struct grid *src, u_int sy, u_int ny)
+grid_duplicate_lines(struct grid *dst, u_int dy, struct grid *src, u_int sy,
+    u_int ny)
 {
 	struct grid_line	*dstl, *srcl;
 	u_int			 yy;
@@ -747,7 +749,7 @@ grid_reflow_split(struct grid *dst, u_int *py, struct grid_line *src_gl,
 		dst_gl->flags |= GRID_LINE_WRAPPED;
 
 		/* Copy the data. */
-		memcpy (&dst_gl->celldata[0], &src_gl->celldata[offset],
+		memcpy(&dst_gl->celldata[0], &src_gl->celldata[offset],
 		    to_copy * sizeof dst_gl->celldata[0]);
 
 		/* Move offset and reduce old line size. */
