@@ -75,6 +75,11 @@ const struct options_table_entry server_options_table[] = {
 	  .default_num = 0
 	},
 
+	{ .name = "focus-events",
+	  .type = OPTIONS_TABLE_FLAG,
+	  .default_num = 0
+	},
+
 	{ .name = "quiet",
 	  .type = OPTIONS_TABLE_FLAG,
 	  .default_num = 0 /* overridden in main() */
@@ -90,6 +95,13 @@ const struct options_table_entry server_options_table[] = {
 
 /* Session options. */
 const struct options_table_entry session_options_table[] = {
+	{ .name = "assume-paste-time",
+	  .type = OPTIONS_TABLE_NUMBER,
+	  .minimum = 0,
+	  .maximum = INT_MAX,
+	  .default_num = 1,
+	},
+
 	{ .name = "base-index",
 	  .type = OPTIONS_TABLE_NUMBER,
 	  .minimum = 0,
@@ -113,11 +125,6 @@ const struct options_table_entry session_options_table[] = {
 	  .default_str = ""
 	},
 
-	{ .name = "default-path",
-	  .type = OPTIONS_TABLE_STRING,
-	  .default_str = ""
-	},
-
 	{ .name = "default-shell",
 	  .type = OPTIONS_TABLE_STRING,
 	  .default_str = _PATH_BSHELL
@@ -188,32 +195,43 @@ const struct options_table_entry session_options_table[] = {
 
 	{ .name = "message-attr",
 	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "message-style"
 	},
 
 	{ .name = "message-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 3
+	  .default_num = 3,
+	  .style = "message-style"
 	},
 
 	{ .name = "message-command-attr",
 	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "message-command-style"
 	},
 
 	{ .name = "message-command-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "message-command-style"
 	},
 
 	{ .name = "message-command-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 3
+	  .default_num = 3,
+	  .style = "message-command-style"
+	},
+
+	{ .name = "message-command-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "bg=black,fg=yellow"
 	},
 
 	{ .name = "message-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "message-style"
 	},
 
 	{ .name = "message-limit",
@@ -223,6 +241,11 @@ const struct options_table_entry session_options_table[] = {
 	  .default_num = 20
 	},
 
+	{ .name = "message-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "bg=yellow,fg=black"
+	},
+
 	{ .name = "mouse-resize-pane",
 	  .type = OPTIONS_TABLE_FLAG,
 	  .default_num = 0
@@ -245,22 +268,36 @@ const struct options_table_entry session_options_table[] = {
 
 	{ .name = "pane-active-border-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "pane-active-border-style"
 	},
 
 	{ .name = "pane-active-border-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 2
+	  .default_num = 2,
+	  .style = "pane-active-border-style"
+	},
+
+	{ .name = "pane-active-border-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "fg=green"
 	},
 
 	{ .name = "pane-border-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "pane-border-style"
 	},
 
 	{ .name = "pane-border-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "pane-border-style"
+	},
+
+	{ .name = "pane-border-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "default"
 	},
 
 	{ .name = "prefix",
@@ -307,17 +344,20 @@ const struct options_table_entry session_options_table[] = {
 
 	{ .name = "status-attr",
 	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "status-style"
 	},
 
 	{ .name = "status-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 2
+	  .default_num = 2,
+	  .style = "status-style"
 	},
 
 	{ .name = "status-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "status-style"
 	},
 
 	{ .name = "status-interval",
@@ -346,17 +386,20 @@ const struct options_table_entry session_options_table[] = {
 
 	{ .name = "status-left-attr",
 	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "status-left-style"
 	},
 
 	{ .name = "status-left-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "status-left-style"
 	},
 
 	{ .name = "status-left-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "status-left-style"
 	},
 
 	{ .name = "status-left-length",
@@ -366,6 +409,11 @@ const struct options_table_entry session_options_table[] = {
 	  .default_num = 10
 	},
 
+	{ .name = "status-left-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "default"
+	},
+
 	{ .name = "status-position",
 	  .type = OPTIONS_TABLE_CHOICE,
 	  .choices = options_table_status_position_list,
@@ -374,22 +422,25 @@ const struct options_table_entry session_options_table[] = {
 
 	{ .name = "status-right",
 	  .type = OPTIONS_TABLE_STRING,
-	  .default_str = "\"#22T\" %H:%M %d-%b-%y"
+	  .default_str = "\"#{=22:pane_title}\" %H:%M %d-%b-%y"
 	},
 
 	{ .name = "status-right-attr",
 	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "status-right-style"
 	},
 
 	{ .name = "status-right-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "status-right-style"
 	},
 
 	{ .name = "status-right-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "status-right-style"
 	},
 
 	{ .name = "status-right-length",
@@ -399,6 +450,16 @@ const struct options_table_entry session_options_table[] = {
 	  .default_num = 40
 	},
 
+	{ .name = "status-right-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "default"
+	},
+
+	{ .name = "status-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "bg=green,fg=black"
+	},
+
 	{ .name = "status-utf8",
 	  .type = OPTIONS_TABLE_FLAG,
 	  .default_num = 0 /* overridden in main() */
@@ -406,10 +467,10 @@ const struct options_table_entry session_options_table[] = {
 
 	{ .name = "terminal-overrides",
 	  .type = OPTIONS_TABLE_STRING,
-	  .default_str = "*88col*:colors=88,*256col*:colors=256"
+	  .default_str = "*256col*:colors=256"
 	                 ",xterm*:XT:Ms=\\E]52;%p1%s;%p2%s\\007"
-	                 ":Cc=\\E]12;%p1%s\\007:Cr=\\E]112\\007"
-			 ":Cs=\\E[%p1%d q:Csr=\\E[2 q,screen*:XT"
+	                 ":Cs=\\E]12;%p1%s\\007:Cr=\\E]112\\007"
+			 ":Ss=\\E[%p1%d q:Se=\\E[2 q,screen*:XT"
 	},
 
 	{ .name = "update-environment",
@@ -469,6 +530,10 @@ const struct options_table_entry window_options_table[] = {
 	  .default_num = 1
 	},
 
+	{ .name = "automatic-rename-format",
+	  .type = OPTIONS_TABLE_STRING,
+	  .default_str = "#{?pane_in_mode,[tmux],#{pane_current_command}}#{?pane_dead,[dead],}"
+	},
 
 	{ .name = "c0-change-trigger",
 	  .type = OPTIONS_TABLE_NUMBER,
@@ -509,13 +574,6 @@ const struct options_table_entry window_options_table[] = {
 	  .default_num = 0
 	},
 
-	{ .name = "layout-history-limit",
-	  .type = OPTIONS_TABLE_NUMBER,
-	  .minimum = 1,
-	  .maximum = USHRT_MAX,
-	  .default_num = 20
-	},
-
 	{ .name = "main-pane-height",
 	  .type = OPTIONS_TABLE_NUMBER,
 	  .minimum = 1,
@@ -532,17 +590,20 @@ const struct options_table_entry window_options_table[] = {
 
 	{ .name = "mode-attr",
 	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "mode-style"
 	},
 
 	{ .name = "mode-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 3
+	  .default_num = 3,
+	  .style = "mode-style"
 	},
 
 	{ .name = "mode-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "mode-style"
 	},
 
 	{ .name = "mode-keys",
@@ -557,6 +618,11 @@ const struct options_table_entry window_options_table[] = {
 	  .default_num = 0
 	},
 
+	{ .name = "mode-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "bg=yellow,fg=black"
+	},
+
 	{ .name = "monitor-activity",
 	  .type = OPTIONS_TABLE_FLAG,
 	  .default_num = 0
@@ -612,72 +678,101 @@ const struct options_table_entry window_options_table[] = {
 
 	{ .name = "window-status-activity-attr",
 	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = GRID_ATTR_REVERSE
+	  .default_num = GRID_ATTR_REVERSE,
+	  .style = "window-status-activity-style"
 	},
 
 	{ .name = "window-status-activity-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "window-status-activity-style"
 	},
 
 	{ .name = "window-status-activity-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "window-status-activity-style"
 	},
 
-	{ .name = "window-status-bell-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = GRID_ATTR_REVERSE
-	},
-
-	{ .name = "window-status-bell-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
-	},
-
-	{ .name = "window-status-bell-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
-	},
-
-	{ .name = "window-status-content-attr",
-	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = GRID_ATTR_REVERSE
-	},
-
-	{ .name = "window-status-content-bg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
-	},
-
-	{ .name = "window-status-content-fg",
-	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	{ .name = "window-status-activity-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "reverse"
 	},
 
 	{ .name = "window-status-attr",
 	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "window-status-style"
+	},
+
+	{ .name = "window-status-bell-attr",
+	  .type = OPTIONS_TABLE_ATTRIBUTES,
+	  .default_num = GRID_ATTR_REVERSE,
+	  .style = "window-status-bell-style"
+	},
+
+	{ .name = "window-status-bell-bg",
+	  .type = OPTIONS_TABLE_COLOUR,
+	  .default_num = 8,
+	  .style = "window-status-bell-style"
+	},
+
+	{ .name = "window-status-bell-fg",
+	  .type = OPTIONS_TABLE_COLOUR,
+	  .default_num = 8,
+	  .style = "window-status-bell-style"
+	},
+
+	{ .name = "window-status-bell-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "reverse"
 	},
 
 	{ .name = "window-status-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "window-status-style"
+	},
+
+	{ .name = "window-status-content-attr",
+	  .type = OPTIONS_TABLE_ATTRIBUTES,
+	  .default_num = GRID_ATTR_REVERSE,
+	  .style = "window-status-content-style"
+	},
+
+	{ .name = "window-status-content-bg",
+	  .type = OPTIONS_TABLE_COLOUR,
+	  .default_num = 8,
+	  .style = "window-status-content-style"
+	},
+
+	{ .name = "window-status-content-fg",
+	  .type = OPTIONS_TABLE_COLOUR,
+	  .default_num = 8,
+	  .style = "window-status-content-style"
+	},
+
+	{ .name = "window-status-content-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "reverse"
 	},
 
 	{ .name = "window-status-current-attr",
 	  .type = OPTIONS_TABLE_ATTRIBUTES,
-	  .default_num = 0
+	  .default_num = 0,
+	  .style = "window-status-current-style"
 	},
 
 	{ .name = "window-status-current-bg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "window-status-current-style"
 	},
 
 	{ .name = "window-status-current-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "window-status-current-style"
 	},
 
 	{ .name = "window-status-current-format",
@@ -685,9 +780,15 @@ const struct options_table_entry window_options_table[] = {
 	  .default_str = "#I:#W#F"
 	},
 
+	{ .name = "window-status-current-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "default"
+	},
+
 	{ .name = "window-status-fg",
 	  .type = OPTIONS_TABLE_COLOUR,
-	  .default_num = 8
+	  .default_num = 8,
+	  .style = "window-status-style"
 	},
 
 	{ .name = "window-status-format",
@@ -695,11 +796,39 @@ const struct options_table_entry window_options_table[] = {
 	  .default_str = "#I:#W#F"
 	},
 
+	{ .name = "window-status-last-attr",
+	  .type = OPTIONS_TABLE_ATTRIBUTES,
+	  .default_num = 0,
+	  .style = "window-status-last-style"
+	},
+
+	{ .name = "window-status-last-bg",
+	  .type = OPTIONS_TABLE_COLOUR,
+	  .default_num = 8,
+	  .style = "window-status-last-style"
+	},
+
+	{ .name = "window-status-last-fg",
+	  .type = OPTIONS_TABLE_COLOUR,
+	  .default_num = 8,
+	  .style = "window-status-last-style"
+	},
+
+	{ .name = "window-status-last-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "default"
+	},
+
 	{ .name = "window-status-separator",
 	  .type = OPTIONS_TABLE_STRING,
 	  .default_str = " "
 	},
 
+	{ .name = "window-status-style",
+	  .type = OPTIONS_TABLE_STYLE,
+	  .default_str = "default"
+	},
+
 	{ .name = "wrap-search",
 	  .type = OPTIONS_TABLE_FLAG,
 	  .default_num = 1
@@ -721,17 +850,24 @@ options_table_populate_tree(
 	const struct options_table_entry	*oe;
 
 	for (oe = table; oe->name != NULL; oe++) {
-		if (oe->default_str != NULL)
+		switch (oe->type) {
+		case OPTIONS_TABLE_STRING:
 			options_set_string(oo, oe->name, "%s", oe->default_str);
-		else
+			break;
+		case OPTIONS_TABLE_STYLE:
+			options_set_style(oo, oe->name, oe->default_str, 0);
+			break;
+		default:
 			options_set_number(oo, oe->name, oe->default_num);
+			break;
+		}
 	}
 }
 
 /* Print an option using its type from the table. */
 const char *
-options_table_print_entry(
-    const struct options_table_entry *oe, struct options_entry *o)
+options_table_print_entry(const struct options_table_entry *oe,
+    struct options_entry *o, int no_quotes)
 {
 	static char	 out[BUFSIZ];
 	const char	*s;
@@ -739,13 +875,17 @@ options_table_print_entry(
 	*out = '\0';
 	switch (oe->type) {
 	case OPTIONS_TABLE_STRING:
-		xsnprintf(out, sizeof out, "\"%s\"", o->str);
+		if (no_quotes)
+			xsnprintf(out, sizeof out, "%s", o->str);
+		else
+			xsnprintf(out, sizeof out, "\"%s\"", o->str);
 		break;
 	case OPTIONS_TABLE_NUMBER:
 		xsnprintf(out, sizeof out, "%lld", o->num);
 		break;
 	case OPTIONS_TABLE_KEY:
-		xsnprintf(out, sizeof out, "%s", key_string_lookup_key(o->num));
+		xsnprintf(out, sizeof out, "%s",
+		    key_string_lookup_key(o->num));
 		break;
 	case OPTIONS_TABLE_COLOUR:
 		s = colour_tostring(o->num);
@@ -765,6 +905,10 @@ options_table_print_entry(
 		s = oe->choices[o->num];
 		xsnprintf(out, sizeof out, "%s", s);
 		break;
+	case OPTIONS_TABLE_STYLE:
+		s = style_tostring(&o->style);
+		xsnprintf(out, sizeof out, "%s", s);
+		break;
 	}
 	return (out);
 }
