@@ -84,14 +84,7 @@ osdep_get_cwd(int fd)
 struct event_base *
 osdep_event_init(void)
 {
-	/*
-	 * On Linux, epoll doesn't work on /dev/null (yes, really).
-	 *
-	 * This has been commented because libevent versions up until the very
-	 * latest (1.4 git or 2.0.10) do not handle signals properly when using
-	 * poll or select, causing hangs.
-	 * 
-	 */
-	/* setenv("EVENT_NOEPOLL", "1", 1); */
+	/* On Linux, epoll doesn't work on /dev/null (yes, really). */
+	setenv("EVENT_NOEPOLL", "1", 1);
 	return (event_init());
 }
