@@ -75,15 +75,17 @@ server_window_check_bell(struct session *s, struct winlink *wl)
 	if (s->curw != wl || s->flags & SESSION_UNATTACHED)
 		wl->flags |= WINLINK_BELL;
 	if (s->flags & SESSION_UNATTACHED)
-		return (1);
-	if (s->curw->window == wl->window)
+		return (0);
+	if (s->curw->window == w)
 		w->flags &= ~WINDOW_BELL;
 
 	visual = options_get_number(&s->options, "visual-bell");
 	action = options_get_number(&s->options, "bell-action");
+	if (action == BELL_NONE)
+		return (0);
 	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
 		c = ARRAY_ITEM(&clients, i);
-		if (c == NULL || c->session != s)
+		if (c == NULL || c->session != s || (c->flags & CLIENT_CONTROL))
 			continue;
 		if (!visual) {
 			tty_bell(&c->tty);
@@ -91,10 +93,8 @@ server_window_check_bell(struct session *s, struct winlink *wl)
 		}
 		if (c->session->curw->window == w)
 			status_message_set(c, "Bell in current window");
-		else if (action == BELL_ANY) {
-			status_message_set(c, "Bell in window %u",
-				winlink_find_by_window(&s->windows, w)->idx);
-		}
+		else if (action == BELL_ANY)
+			status_message_set(c, "Bell in window %u", wl->idx);
 	}
 
 	return (1);
@@ -108,7 +108,7 @@ server_window_check_activity(struct session *s, struct winlink *wl)
 	struct window	*w = wl->window;
 	u_int		 i;
 
-	if (s->curw->window == wl->window)
+	if (s->curw->window == w)
 		w->flags &= ~WINDOW_ACTIVITY;
 
 	if (!(w->flags & WINDOW_ACTIVITY) || wl->flags & WINLINK_ACTIVITY)
@@ -128,8 +128,7 @@ server_window_check_activity(struct session *s, struct winlink *wl)
 			c = ARRAY_ITEM(&clients, i);
 			if (c == NULL || c->session != s)
 				continue;
-			status_message_set(c, "Activity in window %u",
-			    winlink_find_by_window(&s->windows, w)->idx);
+			status_message_set(c, "Activity in window %u", wl->idx);
 		}
 	}
 
@@ -180,8 +179,7 @@ server_window_check_silence(struct session *s, struct winlink *wl)
 			c = ARRAY_ITEM(&clients, i);
 			if (c == NULL || c->session != s)
 				continue;
-			status_message_set(c, "Silence in window %u",
-			    winlink_find_by_window(&s->windows, w)->idx);
+			status_message_set(c, "Silence in window %u", wl->idx);
 		}
 	}
 
@@ -223,8 +221,7 @@ server_window_check_content(
 			c = ARRAY_ITEM(&clients, i);
 			if (c == NULL || c->session != s)
 				continue;
-			status_message_set(c, "Content in window %u",
-			    winlink_find_by_window(&s->windows, w)->idx);
+			status_message_set(c, "Content in window %u", wl->idx);
 		}
 	}
 
@@ -240,7 +237,7 @@ ring_bell(struct session *s)
 
 	for (i = 0; i < ARRAY_LENGTH(&clients); i++) {
 		c = ARRAY_ITEM(&clients, i);
-		if (c != NULL && c->session == s)
+		if (c != NULL && c->session == s && !(c->flags & CLIENT_CONTROL))
 			tty_bell(&c->tty);
 	}
 }
