@@ -1,4 +1,4 @@
-/* $Id: utf8.c,v 1.11 2009/10/23 17:21:34 tcunha Exp $ */
+/* $Id$ */
 
 /*
  * Copyright (c) 2008 Nicholas Marriott <nicm@users.sourceforge.net>
@@ -135,7 +135,9 @@ struct utf8_width_entry utf8_width_table[] = {
 	{ 0x0135f, 0x0135f, 0, NULL, NULL },
 	{ 0x01a17, 0x01a18, 0, NULL, NULL },
 	{ 0x006e7, 0x006e8, 0, NULL, NULL },
+#ifndef __APPLE__
 	{ 0x03099, 0x0309a, 0, NULL, NULL },
+#endif
 	{ 0x00b4d, 0x00b4d, 0, NULL, NULL },
 	{ 0x00ce2, 0x00ce3, 0, NULL, NULL },
 	{ 0x00bcd, 0x00bcd, 0, NULL, NULL },
@@ -171,7 +173,6 @@ struct utf8_width_entry utf8_width_table[] = {
 	{ 0x30000, 0x3fffd, 2, NULL, NULL },
 	{ 0x00711, 0x00711, 0, NULL, NULL },
 	{ 0x0fe00, 0x0fe0f, 0, NULL, NULL },
-	{ 0x01160, 0x011ff, 0, NULL, NULL },
 	{ 0x0180b, 0x0180d, 0, NULL, NULL },
 	{ 0x10a3f, 0x10a3f, 0, NULL, NULL },
 	{ 0x00981, 0x00981, 0, NULL, NULL },
@@ -318,6 +319,19 @@ utf8_combine(const struct utf8_data *utf8data)
 	return (value);
 }
 
+/* Split a two-byte UTF-8 character. */
+u_int
+utf8_split2(u_int uc, u_char *ptr)
+{
+	if (uc > 0x7f) {
+		ptr[0] = (uc >> 6) | 0xc0;
+		ptr[1] = (uc & 0x3f) | 0x80;
+		return (2);
+	}
+	ptr[0] = uc;
+	return (1);
+}
+
 /* Lookup width of UTF-8 data in tree. */
 u_int
 utf8_width(const struct utf8_data *utf8data)
