@@ -31,6 +31,8 @@ void	window_choose_key(struct window_pane *, struct session *, int);
 void	window_choose_mouse(
 	    struct window_pane *, struct session *, struct mouse_event *);
 
+void	window_choose_default_callback(struct window_choose_data *);
+
 void	window_choose_fire_callback(
 	    struct window_pane *, struct window_choose_data *);
 void	window_choose_redraw_screen(struct window_pane *);
@@ -42,8 +44,6 @@ void	window_choose_scroll_down(struct window_pane *);
 
 void	window_choose_collapse(struct window_pane *, struct session *);
 void	window_choose_expand(struct window_pane *, struct session *, u_int);
-void	window_choose_collapse_all(struct window_pane *);
-void	window_choose_expand_all(struct window_pane *);
 
 enum window_choose_input_type {
 	WINDOW_CHOOSE_NORMAL = -1,
@@ -74,13 +74,14 @@ struct window_choose_mode_data {
 	char			*input_str;
 
 	void 			(*callbackfn)(struct window_choose_data *);
-	void			(*freefn)(struct window_choose_data *);
 };
 
+void	window_choose_free1(struct window_choose_mode_data *);
 int     window_choose_key_index(struct window_choose_mode_data *, u_int);
 int     window_choose_index_key(struct window_choose_mode_data *, int);
 void	window_choose_prompt_input(enum window_choose_input_type,
 	    const char *, struct window_pane *, int);
+void	window_choose_reset_top(struct window_pane *, u_int);
 
 void
 window_choose_add(struct window_pane *wp, struct window_choose_data *wcd)
@@ -101,24 +102,41 @@ window_choose_add(struct window_pane *wp, struct window_choose_data *wcd)
 }
 
 void
-window_choose_ready(struct window_pane *wp, u_int cur,
-    void (*callbackfn)(struct window_choose_data *),
-    void (*freefn)(struct window_choose_data *))
+window_choose_set_current(struct window_pane *wp, u_int cur)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
 	struct screen			*s = &data->screen;
 
 	data->selected = cur;
-	if (data->selected > screen_size_y(s) - 1)
-		data->top = ARRAY_LENGTH(&data->list) - screen_size_y(s);
+	window_choose_reset_top(wp, screen_size_y(s));
+}
+
+void
+window_choose_reset_top(struct window_pane *wp, u_int sy)
+{
+	struct window_choose_mode_data	*data = wp->modedata;
+
+	data->top = 0;
+	if (data->selected > sy - 1)
+		data->top = data->selected - (sy - 1);
+
+	window_choose_redraw_screen(wp);
+}
+
+void
+window_choose_ready(struct window_pane *wp, u_int cur,
+    void (*callbackfn)(struct window_choose_data *))
+{
+	struct window_choose_mode_data	*data = wp->modedata;
 
 	data->callbackfn = callbackfn;
-	data->freefn = freefn;
+	if (data->callbackfn == NULL)
+		data->callbackfn = window_choose_default_callback;
 
 	ARRAY_CONCAT(&data->old_list, &data->list);
 
+	window_choose_set_current(wp, cur);
 	window_choose_collapse_all(wp);
-	window_choose_redraw_screen(wp);
 }
 
 struct screen *
@@ -131,7 +149,6 @@ window_choose_init(struct window_pane *wp)
 	wp->modedata = data = xmalloc(sizeof *data);
 
 	data->callbackfn = NULL;
-	data->freefn = NULL;
 	data->input_type = WINDOW_CHOOSE_NORMAL;
 	data->input_str = xstrdup("");
 	data->input_prompt = NULL;
@@ -156,35 +173,104 @@ window_choose_init(struct window_pane *wp)
 }
 
 struct window_choose_data *
-window_choose_data_create(struct cmd_ctx *ctx)
+window_choose_data_create(int type, struct client *c, struct session *s)
 {
 	struct window_choose_data	*wcd;
 
 	wcd = xmalloc(sizeof *wcd);
+	wcd->type = type;
+
 	wcd->ft = format_create();
 	wcd->ft_template = NULL;
+
 	wcd->command = NULL;
+
 	wcd->wl = NULL;
-	wcd->tree_session = NULL;
-	wcd->client = ctx->curclient;
-	wcd->session = ctx->curclient->session;
+	wcd->pane_id = -1;
 	wcd->idx = -1;
-	wcd->type = 0;
+
+	wcd->tree_session = NULL;
+
+	wcd->start_client = c;
+	wcd->start_client->references++;
+	wcd->start_session = s;
+	wcd->start_session->references++;
 
 	return (wcd);
 }
 
+void
+window_choose_data_free(struct window_choose_data *wcd)
+{
+	wcd->start_client->references--;
+	wcd->start_session->references--;
+
+	if (wcd->tree_session != NULL)
+		wcd->tree_session->references--;
+
+	free(wcd->ft_template);
+	format_free(wcd->ft);
+
+	free(wcd->command);
+	free(wcd);
+}
+
+void
+window_choose_data_run(struct window_choose_data *cdata)
+{
+	struct cmd_list	*cmdlist;
+	char		*cause;
+
+	/*
+	 * The command template will have already been replaced. But if it's
+	 * NULL, bail here.
+	 */
+	if (cdata->command == NULL)
+		return;
+
+	if (cmd_string_parse(cdata->command, &cmdlist, NULL, 0, &cause) != 0) {
+		if (cause != NULL) {
+			*cause = toupper((u_char) *cause);
+			status_message_set(cdata->start_client, "%s", cause);
+			free(cause);
+		}
+		return;
+	}
+
+	cmdq_run(cdata->start_client->cmdq, cmdlist);
+	cmd_list_free(cmdlist);
+}
+
+void
+window_choose_default_callback(struct window_choose_data *wcd)
+{
+	if (wcd == NULL)
+		return;
+	if (wcd->start_client->flags & CLIENT_DEAD)
+		return;
+
+	window_choose_data_run(wcd);
+}
+
 void
 window_choose_free(struct window_pane *wp)
 {
-	struct window_choose_mode_data	*data = wp->modedata;
+	if (wp->modedata != NULL)
+		window_choose_free1(wp->modedata);
+}
+
+void
+window_choose_free1(struct window_choose_mode_data *data)
+{
 	struct window_choose_mode_item	*item;
 	u_int				 i;
 
+	if (data == NULL)
+		return;
+
 	for (i = 0; i < ARRAY_LENGTH(&data->old_list); i++) {
 		item = &ARRAY_ITEM(&data->old_list, i);
-		if (data->freefn != NULL && item->wcd != NULL)
-			data->freefn(item->wcd);
+		window_choose_data_free(item->wcd);
 		free(item->name);
 	}
 	ARRAY_FREE(&data->list);
@@ -201,27 +287,23 @@ window_choose_resize(struct window_pane *wp, u_int sx, u_int sy)
 	struct window_choose_mode_data	*data = wp->modedata;
 	struct screen			*s = &data->screen;
 
-	data->top = 0;
-	if (data->selected > sy - 1)
-		data->top = data->selected - (sy - 1);
-
-	screen_resize(s, sx, sy);
+	window_choose_reset_top(wp, sy);
+	screen_resize(s, sx, sy, 0);
 	window_choose_redraw_screen(wp);
 }
 
 void
 window_choose_fire_callback(
-	struct window_pane *wp, struct window_choose_data *wcd)
+    struct window_pane *wp, struct window_choose_data *wcd)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
-	const struct window_mode	*oldmode;
 
-	oldmode = wp->mode;
-	wp->mode = NULL;
+	wp->modedata = NULL;
+	window_pane_reset_mode(wp);
 
 	data->callbackfn(wcd);
 
-	wp->mode = oldmode;
+	window_choose_free1(data);
 }
 
 void
@@ -263,8 +345,7 @@ window_choose_collapse(struct window_pane *wp, struct session *s)
 	 * assign the actual result we want to render and copy the new one over
 	 * the top of it.
 	 */
-	for (i = 0; i < ARRAY_LENGTH(&data->list); i++)
-	{
+	for (i = 0; i < ARRAY_LENGTH(&data->list); i++) {
 		item = &ARRAY_ITEM(&data->list, i);
 		wcd = item->wcd;
 
@@ -298,10 +379,11 @@ window_choose_collapse_all(struct window_pane *wp)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
 	struct window_choose_mode_item	*item;
+	struct screen			*scr = &data->screen;
 	struct session			*s, *chosen;
 	u_int				 i;
 
-	chosen = ARRAY_ITEM(&data->list, data->selected).wcd->session;
+	chosen = ARRAY_ITEM(&data->list, data->selected).wcd->start_session;
 
 	RB_FOREACH(s, sessions, &sessions)
 		window_choose_collapse(wp, s);
@@ -316,7 +398,7 @@ window_choose_collapse_all(struct window_pane *wp)
 		if (item->wcd->type & TREE_SESSION)
 			data->selected = i;
 	}
-	window_choose_redraw_screen(wp);
+	window_choose_reset_top(wp, screen_size_y(scr));
 }
 
 void
@@ -324,6 +406,7 @@ window_choose_expand_all(struct window_pane *wp)
 {
 	struct window_choose_mode_data	*data = wp->modedata;
 	struct window_choose_mode_item	*item;
+	struct screen			*scr = &data->screen;
 	struct session			*s;
 	u_int				 i;
 
@@ -339,7 +422,7 @@ window_choose_expand_all(struct window_pane *wp)
 		}
 	}
 
-	window_choose_redraw_screen(wp);
+	window_choose_reset_top(wp, screen_size_y(scr));
 }
 
 void
@@ -403,7 +486,6 @@ window_choose_expand(struct window_pane *wp, struct session *s, u_int pos)
 	}
 }
 
-/* ARGSUSED */
 void
 window_choose_key(struct window_pane *wp, unused struct session *sess, int key)
 {
@@ -418,7 +500,7 @@ window_choose_key(struct window_pane *wp, unused struct session *sess, int key)
 	items = ARRAY_LENGTH(&data->list);
 
 	if (data->input_type == WINDOW_CHOOSE_GOTO_ITEM) {
-		switch (mode_key_lookup(&data->mdata, key)) {
+		switch (mode_key_lookup(&data->mdata, key, NULL)) {
 		case MODEKEYCHOICE_CANCEL:
 			data->input_type = WINDOW_CHOOSE_NORMAL;
 			window_choose_redraw_screen(wp);
@@ -432,7 +514,6 @@ window_choose_key(struct window_pane *wp, unused struct session *sess, int key)
 			}
 			item = &ARRAY_ITEM(&data->list, n);
 			window_choose_fire_callback(wp, item->wcd);
-			window_pane_reset_mode(wp);
 			break;
 		case MODEKEYCHOICE_BACKSPACE:
 			input_len = strlen(data->input_str);
@@ -450,15 +531,13 @@ window_choose_key(struct window_pane *wp, unused struct session *sess, int key)
 		return;
 	}
 
-	switch (mode_key_lookup(&data->mdata, key)) {
+	switch (mode_key_lookup(&data->mdata, key, NULL)) {
 	case MODEKEYCHOICE_CANCEL:
 		window_choose_fire_callback(wp, NULL);
-		window_pane_reset_mode(wp);
 		break;
 	case MODEKEYCHOICE_CHOOSE:
 		item = &ARRAY_ITEM(&data->list, data->selected);
 		window_choose_fire_callback(wp, item->wcd);
-		window_pane_reset_mode(wp);
 		break;
 	case MODEKEYCHOICE_TREE_TOGGLE:
 		item = &ARRAY_ITEM(&data->list, data->selected);
@@ -608,12 +687,10 @@ window_choose_key(struct window_pane *wp, unused struct session *sess, int key)
 
 		item = &ARRAY_ITEM(&data->list, data->selected);
 		window_choose_fire_callback(wp, item->wcd);
-		window_pane_reset_mode(wp);
 		break;
 	}
 }
 
-/* ARGSUSED */
 void
 window_choose_mouse(
     struct window_pane *wp, unused struct session *sess, struct mouse_event *m)
@@ -623,7 +700,7 @@ window_choose_mouse(
 	struct window_choose_mode_item	*item;
 	u_int				 idx;
 
-	if ((m->b & 3) == 3)
+	if (~m->event & MOUSE_EVENT_CLICK)
 		return;
 	if (m->x >= screen_size_x(s))
 		return;
@@ -637,7 +714,6 @@ window_choose_mouse(
 
 	item = &ARRAY_ITEM(&data->list, data->selected);
 	window_choose_fire_callback(wp, item->wcd);
-	window_pane_reset_mode(wp);
 }
 
 void
@@ -660,7 +736,7 @@ window_choose_write_line(
 	utf8flag = options_get_number(&wp->window->options, "utf8");
 	memcpy(&gc, &grid_default_cell, sizeof gc);
 	if (data->selected == data->top + py)
-		window_mode_attrs(&gc, oo);
+		style_apply(&gc, oo, "mode-style");
 
 	screen_write_cursormove(ctx, 0, py);
 	if (data->top + py  < ARRAY_LENGTH(&data->list)) {
@@ -683,11 +759,11 @@ window_choose_write_line(
 		    (item->wcd->type & TREE_SESSION) ?
 		    (item->state & TREE_EXPANDED ? "-" : "+") : "", item->name);
 	}
-	while (s->cx < screen_size_x(s))
+	while (s->cx < screen_size_x(s) - 1)
 		screen_write_putc(ctx, &gc, ' ');
 
 	if (data->input_type != WINDOW_CHOOSE_NORMAL) {
-		window_mode_attrs(&gc, oo);
+		style_apply(&gc, oo, "mode-style");
 
 		xoff = xsnprintf(hdr, sizeof hdr,
 			"%s: %s", data->input_prompt, data->input_str);
@@ -709,7 +785,7 @@ window_choose_key_index(struct window_choose_mode_data *data, u_int idx)
 	int			mkey;
 
 	for (ptr = keys; *ptr != '\0'; ptr++) {
-		mkey = mode_key_lookup(&data->mdata, *ptr);
+		mkey = mode_key_lookup(&data->mdata, *ptr, NULL);
 		if (mkey != MODEKEY_NONE && mkey != MODEKEY_OTHER)
 			continue;
 		if (idx-- == 0)
@@ -729,7 +805,7 @@ window_choose_index_key(struct window_choose_mode_data *data, int key)
 	u_int			idx = 0;
 
 	for (ptr = keys; *ptr != '\0'; ptr++) {
-		mkey = mode_key_lookup(&data->mdata, *ptr);
+		mkey = mode_key_lookup(&data->mdata, *ptr, NULL);
 		if (mkey != MODEKEY_NONE && mkey != MODEKEY_OTHER)
 			continue;
 		if (key == *ptr)
@@ -792,58 +868,23 @@ window_choose_scroll_down(struct window_pane *wp)
 	screen_write_stop(&ctx);
 }
 
-void
-window_choose_ctx(struct window_choose_data *cdata)
-{
-	struct cmd_ctx		 ctx;
-	struct cmd_list		*cmdlist;
-	char			*cause;
-
-	/* The command template will have already been replaced.  But if it's
-	 * NULL, bail here.
-	 */
-	if (cdata->command == NULL)
-		return;
-
-	if (cmd_string_parse(cdata->command, &cmdlist, &cause) != 0) {
-		if (cause != NULL) {
-			*cause = toupper((u_char) *cause);
-			status_message_set(cdata->client, "%s", cause);
-			free(cause);
-		}
-		return;
-	}
-
-	ctx.msgdata = NULL;
-	ctx.curclient = cdata->client;
-
-	ctx.error = key_bindings_error;
-	ctx.print = key_bindings_print;
-	ctx.info = key_bindings_info;
-
-	ctx.cmdclient = NULL;
-
-	cmd_list_exec(cmdlist, &ctx);
-	cmd_list_free(cmdlist);
-}
-
 struct window_choose_data *
-window_choose_add_session(struct window_pane *wp, struct cmd_ctx *ctx,
-    struct session *s, const char *template, char *action, u_int idx)
+window_choose_add_session(struct window_pane *wp, struct client *c,
+    struct session *s, const char *template, const char *action, u_int idx)
 {
 	struct window_choose_data	*wcd;
 
-	wcd = window_choose_data_create(ctx);
-	wcd->idx = s->idx;
+	wcd = window_choose_data_create(TREE_SESSION, c, c->session);
+	wcd->idx = s->id;
+
 	wcd->tree_session = s;
-	wcd->type = TREE_SESSION;
-	wcd->command = cmd_template_replace(action, s->name, 1);
+	wcd->tree_session->references++;
+
 	wcd->ft_template = xstrdup(template);
 	format_add(wcd->ft, "line", "%u", idx);
 	format_session(wcd->ft, s);
 
-	wcd->client->references++;
-	wcd->session->references++;
+	wcd->command = cmd_template_replace(action, s->name, 1);
 
 	window_choose_add(wp, wcd);
 
@@ -851,63 +892,60 @@ window_choose_add_session(struct window_pane *wp, struct cmd_ctx *ctx,
 }
 
 struct window_choose_data *
-window_choose_add_item(struct window_pane *wp, struct cmd_ctx *ctx,
-    struct winlink *wl, const char *template, char *action, u_int idx)
+window_choose_add_item(struct window_pane *wp, struct client *c,
+    struct winlink *wl, const char *template, const char *action, u_int idx)
 {
 	struct window_choose_data	*wcd;
-	char				*action_data;
+	char				*expanded;
 
-	wcd = window_choose_data_create(ctx);
+	wcd = window_choose_data_create(TREE_OTHER, c, c->session);
 	wcd->idx = wl->idx;
+
 	wcd->ft_template = xstrdup(template);
 	format_add(wcd->ft, "line", "%u", idx);
-	format_session(wcd->ft, wcd->session);
-	format_winlink(wcd->ft, wcd->session, wl);
+	format_session(wcd->ft, wcd->start_session);
+	format_winlink(wcd->ft, wcd->start_session, wl);
 	format_window_pane(wcd->ft, wl->window->active);
 
-	wcd->client->references++;
-	wcd->session->references++;
+	/*
+	 * Interpolate action here, since the data we pass back is the expanded
+	 * template itself.
+	 */
+	xasprintf(&expanded, "%s", format_expand(wcd->ft, wcd->ft_template));
+	wcd->command = cmd_template_replace(action, expanded, 1);
+	free(expanded);
 
 	window_choose_add(wp, wcd);
 
-	/*
-	 * Interpolate action_data here, since the data we pass back is the
-	 * expanded template itself.
-	 */
-	xasprintf(&action_data, "%s", format_expand(wcd->ft, wcd->ft_template));
-	wcd->command = cmd_template_replace(action, action_data, 1);
-	free(action_data);
-
 	return (wcd);
 
 }
 
 struct window_choose_data *
-window_choose_add_window(struct window_pane *wp, struct cmd_ctx *ctx,
+window_choose_add_window(struct window_pane *wp, struct client *c,
     struct session *s, struct winlink *wl, const char *template,
-    char *action, u_int idx)
+    const char *action, u_int idx)
 {
 	struct window_choose_data	*wcd;
-	char				*action_data;
-
-	wcd = window_choose_data_create(ctx);
-
-	xasprintf(&action_data, "%s:%d", s->name, wl->idx);
-	wcd->command = cmd_template_replace(action, action_data, 1);
-	free(action_data);
+	char				*expanded;
 
+	wcd = window_choose_data_create(TREE_WINDOW, c, c->session);
 	wcd->idx = wl->idx;
+
 	wcd->wl = wl;
+
 	wcd->tree_session = s;
-	wcd->type = TREE_WINDOW;
+	wcd->tree_session->references++;
+
 	wcd->ft_template = xstrdup(template);
 	format_add(wcd->ft, "line", "%u", idx);
 	format_session(wcd->ft, s);
 	format_winlink(wcd->ft, s, wl);
 	format_window_pane(wcd->ft, wl->window->active);
 
-	wcd->client->references++;
-	wcd->session->references++;
+	xasprintf(&expanded, "%s:%d", s->name, wl->idx);
+	wcd->command = cmd_template_replace(action, expanded, 1);
+	free(expanded);
 
 	window_choose_add(wp, wcd);
 
