@@ -18,6 +18,7 @@
 
 #include <sys/types.h>
 
+#include <ctype.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -41,17 +42,21 @@ void	window_copy_write_lines(
 
 void	window_copy_scroll_to(struct window_pane *, u_int, u_int);
 int	window_copy_search_compare(
-	    struct grid *, u_int, u_int, struct grid *, u_int);
+	    struct grid *, u_int, u_int, struct grid *, u_int, int);
 int	window_copy_search_lr(
-	    struct grid *, struct grid *, u_int *, u_int, u_int, u_int);
+	    struct grid *, struct grid *, u_int *, u_int, u_int, u_int, int);
 int	window_copy_search_rl(
-	    struct grid *, struct grid *, u_int *, u_int, u_int, u_int);
+	    struct grid *, struct grid *, u_int *, u_int, u_int, u_int, int);
 void	window_copy_search_up(struct window_pane *, const char *);
 void	window_copy_search_down(struct window_pane *, const char *);
 void	window_copy_goto_line(struct window_pane *, const char *);
 void	window_copy_update_cursor(struct window_pane *, u_int, u_int);
 void	window_copy_start_selection(struct window_pane *);
-int	window_copy_update_selection(struct window_pane *);
+int	window_copy_update_selection(struct window_pane *, int);
+void   *window_copy_get_selection(struct window_pane *, size_t *);
+void	window_copy_copy_buffer(struct window_pane *, int, void *, size_t);
+void	window_copy_copy_pipe(
+	    struct window_pane *, struct session *, int, const char *);
 void	window_copy_copy_selection(struct window_pane *, int);
 void	window_copy_clear_selection(struct window_pane *);
 void	window_copy_copy_line(
@@ -61,6 +66,7 @@ u_int	window_copy_find_length(struct window_pane *, u_int);
 void	window_copy_cursor_start_of_line(struct window_pane *);
 void	window_copy_cursor_back_to_indentation(struct window_pane *);
 void	window_copy_cursor_end_of_line(struct window_pane *);
+void	window_copy_other_end(struct window_pane *);
 void	window_copy_cursor_left(struct window_pane *);
 void	window_copy_cursor_right(struct window_pane *);
 void	window_copy_cursor_up(struct window_pane *, int);
@@ -324,7 +330,7 @@ window_copy_pageup(struct window_pane *wp)
 		data->oy = screen_hsize(data->backing);
 	else
 		data->oy += n;
-	window_copy_update_selection(wp);
+	window_copy_update_selection(wp, 1);
 	window_copy_redraw_screen(wp);
 }
 
@@ -335,9 +341,9 @@ window_copy_resize(struct window_pane *wp, u_int sx, u_int sy)
 	struct screen			*s = &data->screen;
 	struct screen_write_ctx	 	 ctx;
 
-	screen_resize(s, sx, sy);
+	screen_resize(s, sx, sy, 0);
 	if (data->backing != &wp->base)
-		screen_resize(data->backing, sx, sy);
+		screen_resize(data->backing, sx, sy, 0);
 
 	if (data->cy > sy - 1)
 		data->cy = sy - 1;
@@ -364,6 +370,7 @@ window_copy_key(struct window_pane *wp, struct session *sess, int key)
 	u_int				 n;
 	int				 np, keys;
 	enum mode_key_cmd		 cmd;
+	const char			*arg;
 
 	np = data->numprefix;
 	if (np <= 0)
@@ -405,11 +412,15 @@ window_copy_key(struct window_pane *wp, struct session *sess, int key)
 		return;
 	}
 
-	cmd = mode_key_lookup(&data->mdata, key);
+	cmd = mode_key_lookup(&data->mdata, key, &arg);
 	switch (cmd) {
 	case MODEKEYCOPY_CANCEL:
 		window_pane_reset_mode(wp);
 		return;
+	case MODEKEYCOPY_OTHEREND:
+		for (; np != 0; np--)
+			window_copy_other_end(wp);
+		break;
 	case MODEKEYCOPY_LEFT:
 		for (; np != 0; np--)
 			window_copy_cursor_left(wp);
@@ -448,7 +459,7 @@ window_copy_key(struct window_pane *wp, struct session *sess, int key)
 			else
 				data->oy -= n;
 		}
-		window_copy_update_selection(wp);
+		window_copy_update_selection(wp, 1);
 		window_copy_redraw_screen(wp);
 		break;
 	case MODEKEYCOPY_HALFPAGEUP:
@@ -459,7 +470,7 @@ window_copy_key(struct window_pane *wp, struct session *sess, int key)
 			else
 				data->oy += n;
 		}
-		window_copy_update_selection(wp);
+		window_copy_update_selection(wp, 1);
 		window_copy_redraw_screen(wp);
 		break;
 	case MODEKEYCOPY_HALFPAGEDOWN:
@@ -470,39 +481,39 @@ window_copy_key(struct window_pane *wp, struct session *sess, int key)
 			else
 				data->oy -= n;
 		}
-		window_copy_update_selection(wp);
+		window_copy_update_selection(wp, 1);
 		window_copy_redraw_screen(wp);
 		break;
 	case MODEKEYCOPY_TOPLINE:
 		data->cx = 0;
 		data->cy = 0;
-		window_copy_update_selection(wp);
+		window_copy_update_selection(wp, 1);
 		window_copy_redraw_screen(wp);
 		break;
 	case MODEKEYCOPY_MIDDLELINE:
 		data->cx = 0;
 		data->cy = (screen_size_y(s) - 1) / 2;
-		window_copy_update_selection(wp);
+		window_copy_update_selection(wp, 1);
 		window_copy_redraw_screen(wp);
 		break;
 	case MODEKEYCOPY_BOTTOMLINE:
 		data->cx = 0;
 		data->cy = screen_size_y(s) - 1;
-		window_copy_update_selection(wp);
+		window_copy_update_selection(wp, 1);
 		window_copy_redraw_screen(wp);
 		break;
 	case MODEKEYCOPY_HISTORYTOP:
 		data->cx = 0;
 		data->cy = 0;
 		data->oy = screen_hsize(data->backing);
-		window_copy_update_selection(wp);
+		window_copy_update_selection(wp, 1);
 		window_copy_redraw_screen(wp);
 		break;
 	case MODEKEYCOPY_HISTORYBOTTOM:
 		data->cx = 0;
 		data->cy = screen_size_y(s) - 1;
 		data->oy = 0;
-		window_copy_update_selection(wp);
+		window_copy_update_selection(wp, 1);
 		window_copy_redraw_screen(wp);
 		break;
 	case MODEKEYCOPY_STARTSELECTION:
@@ -533,6 +544,13 @@ window_copy_key(struct window_pane *wp, struct session *sess, int key)
 		window_copy_clear_selection(wp);
 		window_copy_redraw_screen(wp);
 		break;
+	case MODEKEYCOPY_COPYPIPE:
+		if (sess != NULL) {
+			window_copy_copy_pipe(wp, sess, data->numprefix, arg);
+			window_pane_reset_mode(wp);
+			return;
+		}
+		break;
 	case MODEKEYCOPY_COPYSELECTION:
 		if (sess != NULL) {
 			window_copy_copy_selection(wp, data->numprefix);
@@ -732,10 +750,12 @@ window_copy_key_input(struct window_pane *wp, int key)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*s = &data->screen;
-	size_t				 inputlen;
+	size_t				 inputlen, n;
 	int				 np;
+	struct paste_buffer		*pb;
+	u_char				 ch;
 
-	switch (mode_key_lookup(&data->mdata, key)) {
+	switch (mode_key_lookup(&data->mdata, key, NULL)) {
 	case MODEKEYEDIT_CANCEL:
 		data->numprefix = -1;
 		return (-1);
@@ -747,6 +767,20 @@ window_copy_key_input(struct window_pane *wp, int key)
 	case MODEKEYEDIT_DELETELINE:
 		*data->inputstr = '\0';
 		break;
+	case MODEKEYEDIT_PASTE:
+		if ((pb = paste_get_top(&global_buffers)) == NULL)
+			break;
+		for (n = 0; n < pb->size; n++) {
+			ch = (u_char) pb->data[n];
+			if (ch < 32 || ch == 127)
+				break;
+		}
+		inputlen = strlen(data->inputstr);
+
+		data->inputstr = xrealloc(data->inputstr, 1, inputlen + n + 1);
+		memcpy(data->inputstr + inputlen, pb->data, n);
+		data->inputstr[inputlen + n] = '\0';
+		break;
 	case MODEKEYEDIT_ENTER:
 		np = data->numprefix;
 		if (np <= 0)
@@ -814,7 +848,6 @@ window_copy_key_numeric_prefix(struct window_pane *wp, int key)
 	return (0);
 }
 
-/* ARGSUSED */
 void
 window_copy_mouse(
     struct window_pane *wp, struct session *sess, struct mouse_event *m)
@@ -829,15 +862,19 @@ window_copy_mouse(
 		return;
 
 	/* If mouse wheel (buttons 4 and 5), scroll. */
-	if ((m->b & MOUSE_45)) {
-		if ((m->b & MOUSE_BUTTON) == MOUSE_1) {
+	if (m->event == MOUSE_EVENT_WHEEL) {
+		if (m->wheel == MOUSE_WHEEL_UP) {
 			for (i = 0; i < 5; i++)
-				window_copy_cursor_up(wp, 0);
-		} else if ((m->b & MOUSE_BUTTON) == MOUSE_2) {
+				window_copy_cursor_up(wp, 1);
+		} else if (m->wheel == MOUSE_WHEEL_DOWN) {
 			for (i = 0; i < 5; i++)
-				window_copy_cursor_down(wp, 0);
-			if (data->oy == 0)
-				goto reset_mode;
+				window_copy_cursor_down(wp, 1);
+			/*
+			 * We reached the bottom, leave copy mode,
+			 * but only if no selection is in progress.
+			 */
+			if (data->oy == 0 && !s->sel.flag)
+			    goto reset_mode;
 		}
 		return;
 	}
@@ -847,9 +884,9 @@ window_copy_mouse(
 	 * pressed, or stop the selection on their release.
 	 */
 	if (s->mode & MODE_MOUSE_BUTTON) {
-		if ((m->b & MOUSE_BUTTON) != MOUSE_UP) {
+		if (~m->event & MOUSE_EVENT_UP) {
 			window_copy_update_cursor(wp, m->x, m->y);
-			if (window_copy_update_selection(wp))
+			if (window_copy_update_selection(wp, 1))
 				window_copy_redraw_screen(wp);
 			return;
 		}
@@ -857,7 +894,7 @@ window_copy_mouse(
 	}
 
 	/* Otherwise if other buttons pressed, start selection and motion. */
-	if ((m->b & MOUSE_BUTTON) != MOUSE_UP) {
+	if (~m->event & MOUSE_EVENT_UP) {
 		s->mode &= ~MODE_MOUSE_STANDARD;
 		s->mode |= MODE_MOUSE_BUTTON;
 
@@ -899,47 +936,46 @@ window_copy_scroll_to(struct window_pane *wp, u_int px, u_int py)
 	}
 	data->oy = gd->hsize - offset;
 
-	window_copy_update_selection(wp);
+	window_copy_update_selection(wp, 1);
 	window_copy_redraw_screen(wp);
 }
 
 int
 window_copy_search_compare(
-    struct grid *gd, u_int px, u_int py, struct grid *sgd, u_int spx)
+    struct grid *gd, u_int px, u_int py, struct grid *sgd, u_int spx, int cis)
 {
 	const struct grid_cell	*gc, *sgc;
-	const struct grid_utf8	*gu, *sgu;
+	struct utf8_data	 ud, sud;
 
 	gc = grid_peek_cell(gd, px, py);
+	grid_cell_get(gc, &ud);
 	sgc = grid_peek_cell(sgd, spx, 0);
+	grid_cell_get(sgc, &sud);
 
-	if ((gc->flags & GRID_FLAG_UTF8) != (sgc->flags & GRID_FLAG_UTF8))
+	if (ud.size != sud.size || ud.width != sud.width)
 		return (0);
 
-	if (gc->flags & GRID_FLAG_UTF8) {
-		gu = grid_peek_utf8(gd, px, py);
-		sgu = grid_peek_utf8(sgd, spx, 0);
-		if (grid_utf8_compare(gu, sgu))
-			return (1);
-	} else {
-		if (gc->data == sgc->data)
-			return (1);
-	}
-	return (0);
+	if (cis && ud.size == 1)
+		return (tolower(ud.data[0]) == sud.data[0]);
+
+	return (memcmp(ud.data, sud.data, ud.size) == 0);
 }
 
 int
 window_copy_search_lr(struct grid *gd,
-    struct grid *sgd, u_int *ppx, u_int py, u_int first, u_int last)
+    struct grid *sgd, u_int *ppx, u_int py, u_int first, u_int last, int cis)
 {
 	u_int	ax, bx, px;
+	int	matched;
 
 	for (ax = first; ax < last; ax++) {
 		if (ax + sgd->sx >= gd->sx)
 			break;
 		for (bx = 0; bx < sgd->sx; bx++) {
 			px = ax + bx;
-			if (!window_copy_search_compare(gd, px, py, sgd, bx))
+			matched = window_copy_search_compare(gd, px, py, sgd,
+			    bx, cis);
+			if (!matched)
 				break;
 		}
 		if (bx == sgd->sx) {
@@ -952,16 +988,19 @@ window_copy_search_lr(struct grid *gd,
 
 int
 window_copy_search_rl(struct grid *gd,
-    struct grid *sgd, u_int *ppx, u_int py, u_int first, u_int last)
+    struct grid *sgd, u_int *ppx, u_int py, u_int first, u_int last, int cis)
 {
 	u_int	ax, bx, px;
+	int	matched;
 
 	for (ax = last + 1; ax > first; ax--) {
 		if (gd->sx - (ax - 1) < sgd->sx)
 			continue;
 		for (bx = 0; bx < sgd->sx; bx++) {
 			px = ax - 1 + bx;
-			if (!window_copy_search_compare(gd, px, py, sgd, bx))
+			matched = window_copy_search_compare(gd, px, py, sgd,
+			    bx, cis);
+			if (!matched)
 				break;
 		}
 		if (bx == sgd->sx) {
@@ -982,7 +1021,8 @@ window_copy_search_up(struct window_pane *wp, const char *searchstr)
 	struct grid_cell	 	 gc;
 	size_t				 searchlen;
 	u_int				 i, last, fx, fy, px;
-	int				 utf8flag, n, wrapped, wrapflag;
+	int				 utf8flag, n, wrapped, wrapflag, cis;
+	const char			*ptr;
 
 	if (*searchstr == '\0')
 		return;
@@ -1008,13 +1048,21 @@ window_copy_search_up(struct window_pane *wp, const char *searchstr)
 		fx--;
 	n = wrapped = 0;
 
+	cis = 1;
+	for (ptr = searchstr; *ptr != '\0'; ptr++) {
+		if (*ptr != tolower((u_char)*ptr)) {
+			cis = 0;
+			break;
+		}
+	}
+
 retry:
 	sgd = ss.grid;
 	for (i = fy + 1; i > 0; i--) {
 		last = screen_size_x(s);
 		if (i == fy + 1)
 			last = fx;
-		n = window_copy_search_rl(gd, sgd, &px, i - 1, 0, last);
+		n = window_copy_search_rl(gd, sgd, &px, i - 1, 0, last, cis);
 		if (n) {
 			window_copy_scroll_to(wp, px, i - 1);
 			break;
@@ -1040,7 +1088,8 @@ window_copy_search_down(struct window_pane *wp, const char *searchstr)
 	struct grid_cell	 	 gc;
 	size_t				 searchlen;
 	u_int				 i, first, fx, fy, px;
-	int				 utf8flag, n, wrapped, wrapflag;
+	int				 utf8flag, n, wrapped, wrapflag, cis;
+	const char			*ptr;
 
 	if (*searchstr == '\0')
 		return;
@@ -1066,13 +1115,22 @@ window_copy_search_down(struct window_pane *wp, const char *searchstr)
 		fx++;
 	n = wrapped = 0;
 
+	cis = 1;
+	for (ptr = searchstr; *ptr != '\0'; ptr++) {
+		if (*ptr != tolower((u_char)*ptr)) {
+			cis = 0;
+			break;
+		}
+	}
+
 retry:
 	sgd = ss.grid;
 	for (i = fy + 1; i < gd->hsize + gd->sy + 1; i++) {
 		first = 0;
 		if (i == fy + 1)
 			first = fx;
-		n = window_copy_search_lr(gd, sgd, &px, i - 1, first, gd->sx);
+		n = window_copy_search_lr(gd, sgd, &px, i - 1, first, gd->sx,
+		    cis);
 		if (n) {
 			window_copy_scroll_to(wp, px, i - 1);
 			break;
@@ -1100,7 +1158,7 @@ window_copy_goto_line(struct window_pane *wp, const char *linestr)
 		return;
 
 	data->oy = lineno;
-	window_copy_update_selection(wp);
+	window_copy_update_selection(wp, 1);
 	window_copy_redraw_screen(wp);
 }
 
@@ -1112,10 +1170,10 @@ window_copy_write_line(
 	struct screen			*s = &data->screen;
 	struct options			*oo = &wp->window->options;
 	struct grid_cell		 gc;
-	char				 hdr[32];
-	size_t	 			 last, xoff = 0, size = 0;
+	char				 hdr[512];
+	size_t				 last, xoff = 0, size = 0, limit;
 
-	window_mode_attrs(&gc, oo);
+	style_apply(&gc, oo, "mode-style");
 
 	last = screen_size_y(s) - 1;
 	if (py == 0) {
@@ -1126,11 +1184,14 @@ window_copy_write_line(
 		screen_write_cursormove(ctx, screen_size_x(s) - size, 0);
 		screen_write_puts(ctx, &gc, "%s", hdr);
 	} else if (py == last && data->inputtype != WINDOW_COPY_OFF) {
+		limit = sizeof hdr;
+		if (limit > screen_size_x(s))
+			limit = screen_size_x(s);
 		if (data->inputtype == WINDOW_COPY_NUMERICPREFIX) {
-			xoff = size = xsnprintf(hdr, sizeof hdr,
+			xoff = size = xsnprintf(hdr, limit,
 			    "Repeat: %u", data->numprefix);
 		} else {
-			xoff = size = xsnprintf(hdr, sizeof hdr,
+			xoff = size = xsnprintf(hdr, limit,
 			    "%s: %s", data->inputprompt, data->inputstr);
 		}
 		screen_write_cursormove(ctx, 0, last);
@@ -1213,11 +1274,11 @@ window_copy_start_selection(struct window_pane *wp)
 	data->sely = screen_hsize(data->backing) + data->cy - data->oy;
 
 	s->sel.flag = 1;
-	window_copy_update_selection(wp);
+	window_copy_update_selection(wp, 1);
 }
 
 int
-window_copy_update_selection(struct window_pane *wp)
+window_copy_update_selection(struct window_pane *wp, int may_redraw)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*s = &data->screen;
@@ -1229,7 +1290,7 @@ window_copy_update_selection(struct window_pane *wp)
 		return (0);
 
 	/* Set colours. */
-	window_mode_attrs(&gc, oo);
+	style_apply(&gc, oo, "mode-style");
 
 	/* Find top of screen. */
 	ty = screen_hsize(data->backing) - data->oy;
@@ -1252,7 +1313,7 @@ window_copy_update_selection(struct window_pane *wp)
 	screen_set_selection(s,
 	    sx, sy, data->cx, screen_hsize(s) + data->cy, data->rectflag, &gc);
 
-	if (data->rectflag) {
+	if (data->rectflag && may_redraw) {
 		/*
 		 * Can't rely on the caller to redraw the right lines for
 		 * rectangle selection - find the highest line and the number
@@ -1268,19 +1329,19 @@ window_copy_update_selection(struct window_pane *wp)
 	return (1);
 }
 
-void
-window_copy_copy_selection(struct window_pane *wp, int idx)
+void *
+window_copy_get_selection(struct window_pane *wp, size_t *len)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*s = &data->screen;
 	char				*buf;
 	size_t				 off;
-	u_int				 i, xx, yy, sx, sy, ex, ey, limit;
+	u_int				 i, xx, yy, sx, sy, ex, ey;
 	u_int				 firstsx, lastex, restex, restsx;
 	int				 keys;
 
 	if (!s->sel.flag)
-		return;
+		return (NULL);
 
 	buf = xmalloc(1);
 	off = 0;
@@ -1373,19 +1434,61 @@ window_copy_copy_selection(struct window_pane *wp, int idx)
 	/* Don't bother if no data. */
 	if (off == 0) {
 		free(buf);
-		return;
+		return (NULL);
 	}
-	off--;	/* remove final \n */
+	*len = off - 1;	/* remove final \n */
+	return (buf);
+}
 
-	if (options_get_number(&global_options, "set-clipboard"))
-		screen_write_setselection(&wp->ictx.ctx, buf, off);
+void
+window_copy_copy_buffer(struct window_pane *wp, int idx, void *buf, size_t len)
+{
+	u_int			limit;
+	struct screen_write_ctx	ctx;
+
+	if (options_get_number(&global_options, "set-clipboard")) {
+		screen_write_start(&ctx, wp, NULL);
+		screen_write_setselection(&ctx, buf, len);
+		screen_write_stop(&ctx);
+	}
 
-	/* Add the buffer to the stack. */
 	if (idx == -1) {
 		limit = options_get_number(&global_options, "buffer-limit");
-		paste_add(&global_buffers, buf, off, limit);
-	} else
-		paste_replace(&global_buffers, idx, buf, off);
+		paste_add(&global_buffers, buf, len, limit);
+	} else if (paste_replace(&global_buffers, idx, buf, len) != 0)
+		free(buf);
+}
+
+void
+window_copy_copy_pipe(
+    struct window_pane *wp, struct session *sess, int idx, const char *arg)
+{
+	void		*buf;
+	size_t		 len;
+	struct job	*job;
+
+
+	buf = window_copy_get_selection(wp, &len);
+	if (buf == NULL)
+		return;
+
+	job = job_run(arg, sess, NULL, NULL, NULL);
+	bufferevent_write(job->event, buf, len);
+
+	window_copy_copy_buffer(wp, idx, buf, len);
+}
+
+void
+window_copy_copy_selection(struct window_pane *wp, int idx)
+{
+	void*	buf;
+	size_t	len;
+
+	buf = window_copy_get_selection(wp, &len);
+	if (buf == NULL)
+		return;
+
+	window_copy_copy_buffer(wp, idx, buf, len);
 }
 
 void
@@ -1395,10 +1498,9 @@ window_copy_copy_line(struct window_pane *wp,
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct grid			*gd = data->backing->grid;
 	const struct grid_cell		*gc;
-	const struct grid_utf8		*gu;
 	struct grid_line		*gl;
+	struct utf8_data		 ud;
 	u_int				 i, xx, wrapped = 0;
-	size_t				 size;
 
 	if (sx > ex)
 		return;
@@ -1426,15 +1528,11 @@ window_copy_copy_line(struct window_pane *wp,
 			gc = grid_peek_cell(gd, i, sy);
 			if (gc->flags & GRID_FLAG_PADDING)
 				continue;
-			if (!(gc->flags & GRID_FLAG_UTF8)) {
-				*buf = xrealloc(*buf, 1, (*off) + 1);
-				(*buf)[(*off)++] = gc->data;
-			} else {
-				gu = grid_peek_utf8(gd, i, sy);
-				size = grid_utf8_size(gu);
-				*buf = xrealloc(*buf, 1, (*off) + size);
-				*off += grid_utf8_copy(gu, *buf + *off, size);
-			}
+			grid_cell_get(gc, &ud);
+
+			*buf = xrealloc(*buf, 1, (*off) + ud.size);
+			memcpy(*buf + *off, ud.data, ud.size);
+			*off += ud.size;
 		}
 	}
 
@@ -1464,13 +1562,15 @@ window_copy_in_set(struct window_pane *wp, u_int px, u_int py, const char *set)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
 	const struct grid_cell		*gc;
+	struct utf8_data		 ud;
 
 	gc = grid_peek_cell(data->backing->grid, px, py);
-	if (gc->flags & (GRID_FLAG_PADDING|GRID_FLAG_UTF8))
+	grid_cell_get(gc, &ud);
+	if (ud.size != 1 || gc->flags & GRID_FLAG_PADDING)
 		return (0);
-	if (gc->data == 0x00 || gc->data == 0x7f)
+	if (*ud.data == 0x00 || *ud.data == 0x7f)
 		return (0);
-	return (strchr(set, gc->data) != NULL);
+	return (strchr(set, *ud.data) != NULL);
 }
 
 u_int
@@ -1479,6 +1579,7 @@ window_copy_find_length(struct window_pane *wp, u_int py)
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*s = data->backing;
 	const struct grid_cell		*gc;
+	struct utf8_data		 ud;
 	u_int				 px;
 
 	/*
@@ -1492,9 +1593,8 @@ window_copy_find_length(struct window_pane *wp, u_int py)
 		px = screen_size_x(s);
 	while (px > 0) {
 		gc = grid_peek_cell(s->grid, px - 1, py);
-		if (gc->flags & GRID_FLAG_UTF8)
-			break;
-		if (gc->data != ' ')
+		grid_cell_get(gc, &ud);
+		if (ud.size != 1 || *ud.data != ' ')
 			break;
 		px--;
 	}
@@ -1517,7 +1617,7 @@ window_copy_cursor_start_of_line(struct window_pane *wp)
 		}
 	}
 	window_copy_update_cursor(wp, 0, data->cy);
-	if (window_copy_update_selection(wp))
+	if (window_copy_update_selection(wp, 1))
 		window_copy_redraw_lines(wp, data->cy, 1);
 }
 
@@ -1527,6 +1627,7 @@ window_copy_cursor_back_to_indentation(struct window_pane *wp)
 	struct window_copy_mode_data	*data = wp->modedata;
 	u_int				 px, py, xx;
 	const struct grid_cell		*gc;
+	struct utf8_data		 ud;
 
 	px = 0;
 	py = screen_hsize(data->backing) + data->cy - data->oy;
@@ -1534,15 +1635,14 @@ window_copy_cursor_back_to_indentation(struct window_pane *wp)
 
 	while (px < xx) {
 		gc = grid_peek_cell(data->backing->grid, px, py);
-		if (gc->flags & GRID_FLAG_UTF8)
-			break;
-		if (gc->data != ' ')
+		grid_cell_get(gc, &ud);
+		if (ud.size != 1 || *ud.data != ' ')
 			break;
 		px++;
 	}
 
 	window_copy_update_cursor(wp, px, data->cy);
-	if (window_copy_update_selection(wp))
+	if (window_copy_update_selection(wp, 1))
 		window_copy_redraw_lines(wp, data->cy, 1);
 }
 
@@ -1572,10 +1672,43 @@ window_copy_cursor_end_of_line(struct window_pane *wp)
 	}
 	window_copy_update_cursor(wp, px, data->cy);
 
-	if (window_copy_update_selection(wp))
+	if (window_copy_update_selection(wp, 1))
 		window_copy_redraw_lines(wp, data->cy, 1);
 }
 
+void
+window_copy_other_end(struct window_pane *wp)
+{
+	struct window_copy_mode_data	*data = wp->modedata;
+	struct screen			*s = &data->screen;
+	u_int				 selx, sely, cx, cy, yy;
+
+	if (!s->sel.flag)
+		return;
+
+	selx = data->selx;
+	sely = data->sely;
+	cx = data->cx;
+	cy = data->cy;
+	yy = screen_hsize(data->backing) + data->cy - data->oy;
+
+	data->selx = cx;
+	data->sely = yy;
+	data->cx = selx;
+
+	if (sely < screen_hsize(data->backing) - data->oy) {
+		data->oy = screen_hsize(data->backing) - sely;
+		data->cy = 0;
+	} else if (sely > screen_hsize(data->backing) - data->oy + screen_size_y(s)) {
+		data->oy = screen_hsize(data->backing) - sely + screen_size_y(s) - 1;
+		data->cy = screen_size_y(s) - 1;
+
+	} else
+		data->cy = cy + sely - yy;
+
+	window_copy_redraw_screen(wp);
+}
+
 void
 window_copy_cursor_left(struct window_pane *wp)
 {
@@ -1586,7 +1719,7 @@ window_copy_cursor_left(struct window_pane *wp)
 		window_copy_cursor_end_of_line(wp);
 	} else {
 		window_copy_update_cursor(wp, data->cx - 1, data->cy);
-		if (window_copy_update_selection(wp))
+		if (window_copy_update_selection(wp, 1))
 			window_copy_redraw_lines(wp, data->cy, 1);
 	}
 }
@@ -1609,7 +1742,7 @@ window_copy_cursor_right(struct window_pane *wp)
 		window_copy_cursor_down(wp, 0);
 	} else {
 		window_copy_update_cursor(wp, data->cx + 1, data->cy);
-		if (window_copy_update_selection(wp))
+		if (window_copy_update_selection(wp, 1))
 			window_copy_redraw_lines(wp, data->cy, 1);
 	}
 }
@@ -1639,7 +1772,7 @@ window_copy_cursor_up(struct window_pane *wp, int scroll_only)
 		}
 	} else {
 		window_copy_update_cursor(wp, data->cx, data->cy - 1);
-		if (window_copy_update_selection(wp)) {
+		if (window_copy_update_selection(wp, 1)) {
 			if (data->cy == screen_size_y(s) - 1)
 				window_copy_redraw_lines(wp, data->cy, 1);
 			else
@@ -1677,7 +1810,7 @@ window_copy_cursor_down(struct window_pane *wp, int scroll_only)
 			window_copy_redraw_lines(wp, data->cy - 1, 2);
 	} else {
 		window_copy_update_cursor(wp, data->cx, data->cy + 1);
-		if (window_copy_update_selection(wp))
+		if (window_copy_update_selection(wp, 1))
 			window_copy_redraw_lines(wp, data->cy - 1, 2);
 	}
 
@@ -1696,6 +1829,7 @@ window_copy_cursor_jump(struct window_pane *wp)
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*back_s = data->backing;
 	const struct grid_cell		*gc;
+	struct utf8_data		 ud;
 	u_int				 px, py, xx;
 
 	px = data->cx + 1;
@@ -1704,11 +1838,11 @@ window_copy_cursor_jump(struct window_pane *wp)
 
 	while (px < xx) {
 		gc = grid_peek_cell(back_s->grid, px, py);
-		if ((gc->flags & (GRID_FLAG_PADDING|GRID_FLAG_UTF8)) == 0
-		    && gc->data == data->jumpchar) {
-
+		grid_cell_get(gc, &ud);
+		if (!(gc->flags & GRID_FLAG_PADDING) &&
+		    ud.size == 1 && *ud.data == data->jumpchar) {
 			window_copy_update_cursor(wp, px, data->cy);
-			if (window_copy_update_selection(wp))
+			if (window_copy_update_selection(wp, 1))
 				window_copy_redraw_lines(wp, data->cy, 1);
 			return;
 		}
@@ -1722,6 +1856,7 @@ window_copy_cursor_jump_back(struct window_pane *wp)
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*back_s = data->backing;
 	const struct grid_cell		*gc;
+	struct utf8_data		 ud;
 	u_int				 px, py;
 
 	px = data->cx;
@@ -1732,11 +1867,11 @@ window_copy_cursor_jump_back(struct window_pane *wp)
 
 	for (;;) {
 		gc = grid_peek_cell(back_s->grid, px, py);
-		if ((gc->flags & (GRID_FLAG_PADDING|GRID_FLAG_UTF8)) == 0
-		    && gc->data == data->jumpchar) {
-
+		grid_cell_get(gc, &ud);
+		if (!(gc->flags & GRID_FLAG_PADDING) &&
+		    ud.size == 1 && *ud.data == data->jumpchar) {
 			window_copy_update_cursor(wp, px, data->cy);
-			if (window_copy_update_selection(wp))
+			if (window_copy_update_selection(wp, 1))
 				window_copy_redraw_lines(wp, data->cy, 1);
 			return;
 		}
@@ -1752,6 +1887,7 @@ window_copy_cursor_jump_to(struct window_pane *wp)
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*back_s = data->backing;
 	const struct grid_cell		*gc;
+	struct utf8_data		 ud;
 	u_int				 px, py, xx;
 
 	px = data->cx + 1;
@@ -1760,11 +1896,11 @@ window_copy_cursor_jump_to(struct window_pane *wp)
 
 	while (px < xx) {
 		gc = grid_peek_cell(back_s->grid, px, py);
-		if ((gc->flags & (GRID_FLAG_PADDING|GRID_FLAG_UTF8)) == 0
-		    && gc->data == data->jumpchar) {
-
+		grid_cell_get(gc, &ud);
+		if (!(gc->flags & GRID_FLAG_PADDING) &&
+		    ud.size == 1 && *ud.data == data->jumpchar) {
 			window_copy_update_cursor(wp, px - 1, data->cy);
-			if (window_copy_update_selection(wp))
+			if (window_copy_update_selection(wp, 1))
 				window_copy_redraw_lines(wp, data->cy, 1);
 			return;
 		}
@@ -1778,6 +1914,7 @@ window_copy_cursor_jump_to_back(struct window_pane *wp)
 	struct window_copy_mode_data	*data = wp->modedata;
 	struct screen			*back_s = data->backing;
 	const struct grid_cell		*gc;
+	struct utf8_data		 ud;
 	u_int				 px, py;
 
 	px = data->cx;
@@ -1788,11 +1925,11 @@ window_copy_cursor_jump_to_back(struct window_pane *wp)
 
 	for (;;) {
 		gc = grid_peek_cell(back_s->grid, px, py);
-		if ((gc->flags & (GRID_FLAG_PADDING|GRID_FLAG_UTF8)) == 0
-		    && gc->data == data->jumpchar) {
-
+		grid_cell_get(gc, &ud);
+		if (!(gc->flags & GRID_FLAG_PADDING) &&
+		    ud.size == 1 && *ud.data == data->jumpchar) {
 			window_copy_update_cursor(wp, px + 1, data->cy);
-			if (window_copy_update_selection(wp))
+			if (window_copy_update_selection(wp, 1))
 				window_copy_redraw_lines(wp, data->cy, 1);
 			return;
 		}
@@ -1840,7 +1977,7 @@ window_copy_cursor_next_word(struct window_pane *wp, const char *separators)
 	} while (expected == 1);
 
 	window_copy_update_cursor(wp, px, data->cy);
-	if (window_copy_update_selection(wp))
+	if (window_copy_update_selection(wp, 1))
 		window_copy_redraw_lines(wp, data->cy, 1);
 }
 
@@ -1848,15 +1985,20 @@ void
 window_copy_cursor_next_word_end(struct window_pane *wp, const char *separators)
 {
 	struct window_copy_mode_data	*data = wp->modedata;
+	struct options			*oo = &wp->window->options;
 	struct screen			*back_s = data->backing;
 	u_int				 px, py, xx, yy;
-	int				 expected = 1;
+	int				 keys, expected = 1;
 
 	px = data->cx;
 	py = screen_hsize(back_s) + data->cy - data->oy;
 	xx = window_copy_find_length(wp, py);
 	yy = screen_hsize(back_s) + screen_size_y(back_s) - 1;
 
+	keys = options_get_number(oo, "mode-keys");
+	if (keys == MODEKEY_VI && !window_copy_in_set(wp, px, py, separators))
+		px++;
+
 	/*
 	 * First skip past any word characters, then any nonword characters.
 	 *
@@ -1881,8 +2023,11 @@ window_copy_cursor_next_word_end(struct window_pane *wp, const char *separators)
 		expected = !expected;
 	} while (expected == 0);
 
+	if (keys == MODEKEY_VI && px != 0)
+		px--;
+
 	window_copy_update_cursor(wp, px, data->cy);
-	if (window_copy_update_selection(wp))
+	if (window_copy_update_selection(wp, 1))
 		window_copy_redraw_lines(wp, data->cy, 1);
 }
 
@@ -1920,7 +2065,7 @@ window_copy_cursor_previous_word(struct window_pane *wp, const char *separators)
 
 out:
 	window_copy_update_cursor(wp, px, data->cy);
-	if (window_copy_update_selection(wp))
+	if (window_copy_update_selection(wp, 1))
 		window_copy_redraw_lines(wp, data->cy, 1);
 }
 
@@ -1937,6 +2082,8 @@ window_copy_scroll_up(struct window_pane *wp, u_int ny)
 		return;
 	data->oy -= ny;
 
+	window_copy_update_selection(wp, 0);
+
 	screen_write_start(&ctx, wp, NULL);
 	screen_write_cursormove(&ctx, 0, 0);
 	screen_write_deleteline(&ctx, ny);
@@ -1946,12 +2093,9 @@ window_copy_scroll_up(struct window_pane *wp, u_int ny)
 		window_copy_write_line(wp, &ctx, 1);
 	if (screen_size_y(s) > 3)
 		window_copy_write_line(wp, &ctx, screen_size_y(s) - 2);
-	if (s->sel.flag && screen_size_y(s) > ny) {
-		window_copy_update_selection(wp);
+	if (s->sel.flag && screen_size_y(s) > ny)
 		window_copy_write_line(wp, &ctx, screen_size_y(s) - ny - 1);
-	}
 	screen_write_cursormove(&ctx, data->cx, data->cy);
-	window_copy_update_selection(wp);
 	screen_write_stop(&ctx);
 }
 
@@ -1971,17 +2115,17 @@ window_copy_scroll_down(struct window_pane *wp, u_int ny)
 		return;
 	data->oy += ny;
 
+	window_copy_update_selection(wp, 0);
+
 	screen_write_start(&ctx, wp, NULL);
 	screen_write_cursormove(&ctx, 0, 0);
 	screen_write_insertline(&ctx, ny);
 	window_copy_write_lines(wp, &ctx, 0, ny);
-	if (s->sel.flag && screen_size_y(s) > ny) {
-		window_copy_update_selection(wp);
+	if (s->sel.flag && screen_size_y(s) > ny)
 		window_copy_write_line(wp, &ctx, ny);
-	} else if (ny == 1) /* nuke position */
+	else if (ny == 1) /* nuke position */
 		window_copy_write_line(wp, &ctx, 1);
 	screen_write_cursormove(&ctx, data->cx, data->cy);
-	window_copy_update_selection(wp);
 	screen_write_stop(&ctx);
 }
 
@@ -1998,6 +2142,6 @@ window_copy_rectangle_toggle(struct window_pane *wp)
 	if (data->cx > px)
 		window_copy_update_cursor(wp, px, data->cy);
 
-	window_copy_update_selection(wp);
+	window_copy_update_selection(wp, 1);
 	window_copy_redraw_screen(wp);
 }
