@@ -58,6 +58,10 @@ struct window_pane_tree all_window_panes;
 u_int	next_window_pane_id;
 u_int	next_window_id;
 
+struct window_pane *window_pane_active_set(struct window_pane *,
+	    struct window_pane *);
+void	window_pane_active_lost(struct window_pane *, struct window_pane *);
+
 void	window_pane_timer_callback(int, short, void *);
 void	window_pane_read_callback(struct bufferevent *, void *);
 void	window_pane_error_callback(struct bufferevent *, short, void *);
@@ -306,7 +310,7 @@ window_create1(u_int sx, u_int sy)
 
 struct window *
 window_create(const char *name, const char *cmd, const char *shell,
-    const char *cwd, struct environ *env, struct termios *tio,
+    int cwd, struct environ *env, struct termios *tio,
     u_int sx, u_int sy, u_int hlimit, char **cause)
 {
 	struct window		*w;
@@ -383,6 +387,59 @@ window_resize(struct window *w, u_int sx, u_int sy)
 	w->sy = sy;
 }
 
+/*
+ * Restore previously active pane when changing from wp to nextwp. The intended
+ * pane is in nextwp and it returns the previously focused pane.
+ */
+struct window_pane *
+window_pane_active_set(struct window_pane *wp, struct window_pane *nextwp)
+{
+	struct layout_cell	*lc;
+	struct window_pane	*lastwp;
+
+	/* Target pane's parent must not be an ancestor of source pane. */
+	for (lc = wp->layout_cell->parent; lc != NULL; lc = lc->parent) {
+		if (lc == nextwp->layout_cell->parent)
+			return (nextwp);
+	}
+
+	/*
+	 * Previously active pane, if any, must not be the same as the source
+	 * pane.
+	 */
+	if (nextwp->layout_cell->parent != NULL) {
+		lastwp = nextwp->layout_cell->parent->lastwp;
+		if (lastwp != wp && window_pane_visible(lastwp))
+			return (lastwp);
+	}
+	return (nextwp);
+}
+
+/* Remember previously active pane when changing from wp to nextwp. */
+void
+window_pane_active_lost(struct window_pane *wp, struct window_pane *nextwp)
+{
+	struct layout_cell	*lc, *lc2;
+
+	/* Save the target pane in its parent. */
+	nextwp->layout_cell->parent->lastwp = nextwp;
+
+	/*
+	 * Save the source pane in all of its parents up to, but not including,
+	 * the common ancestor of itself and the target panes.
+	 */
+	if (wp == NULL)
+		return;
+	for (lc = wp->layout_cell->parent; lc != NULL; lc = lc->parent) {
+		lc2 = nextwp->layout_cell->parent;
+		for (; lc2 != NULL; lc2 = lc2->parent) {
+			if (lc == lc2)
+				return;
+		}
+		lc->lastwp = wp;
+	}
+}
+
 void
 window_set_active_pane(struct window *w, struct window_pane *wp)
 {
@@ -390,6 +447,7 @@ window_set_active_pane(struct window *w, struct window_pane *wp)
 		return;
 	w->last = w->active;
 	w->active = wp;
+	window_pane_active_lost(w->last, wp);
 	while (!window_pane_visible(w->active)) {
 		w->active = TAILQ_PREV(w->active, window_panes, entry);
 		if (w->active == NULL)
@@ -672,7 +730,7 @@ window_pane_create(struct window *w, u_int sx, u_int sy, u_int hlimit)
 
 	wp->cmd = NULL;
 	wp->shell = NULL;
-	wp->cwd = NULL;
+	wp->cwd = -1;
 
 	wp->fd = -1;
 	wp->event = NULL;
@@ -704,6 +762,16 @@ window_pane_create(struct window *w, u_int sx, u_int sy, u_int hlimit)
 void
 window_pane_destroy(struct window_pane *wp)
 {
+	struct window_pane	*wp2;
+
+	/* Forget removed pane in all layout cells that remember it. */
+	RB_FOREACH(wp2, window_pane_tree, &all_window_panes) {
+		if (wp2->layout_cell != NULL &&
+		    wp2->layout_cell->parent != NULL &&
+		    wp2->layout_cell->parent->lastwp == wp)
+			wp2->layout_cell->parent->lastwp = NULL;
+	}
+
 	window_pane_reset_mode(wp);
 
 	if (event_initialized(&wp->changes_timer))
@@ -727,7 +795,7 @@ window_pane_destroy(struct window_pane *wp)
 
 	RB_REMOVE(window_pane_tree, &all_window_panes, wp);
 
-	free(wp->cwd);
+	close(wp->cwd);
 	free(wp->shell);
 	free(wp->cmd);
 	free(wp);
@@ -735,7 +803,7 @@ window_pane_destroy(struct window_pane *wp)
 
 int
 window_pane_spawn(struct window_pane *wp, const char *cmd, const char *shell,
-    const char *cwd, struct environ *env, struct termios *tio, char **cause)
+    int cwd, struct environ *env, struct termios *tio, char **cause)
 {
 	struct winsize	 ws;
 	char		*argv0, paneid[16];
@@ -754,9 +822,9 @@ window_pane_spawn(struct window_pane *wp, const char *cmd, const char *shell,
 		free(wp->shell);
 		wp->shell = xstrdup(shell);
 	}
-	if (cwd != NULL) {
-		free(wp->cwd);
-		wp->cwd = xstrdup(cwd);
+	if (cwd != -1) {
+		close(wp->cwd);
+		wp->cwd = dup(cwd);
 	}
 
 	log_debug("spawn: %s -- %s", wp->shell, wp->cmd);
@@ -771,7 +839,7 @@ window_pane_spawn(struct window_pane *wp, const char *cmd, const char *shell,
 		xasprintf(cause, "%s: %s", cmd, strerror(errno));
 		return (-1);
 	case 0:
-		if (chdir(wp->cwd) != 0)
+		if (fchdir(wp->cwd) != 0)
 			chdir("/");
 
 		if (tcgetattr(STDIN_FILENO, &tio2) != 0)
@@ -1131,7 +1199,7 @@ window_pane_find_up(struct window_pane *wp)
 		if (wp2->yoff + wp2->sy + 1 != top)
 			continue;
 		if (left >= wp2->xoff && left <= wp2->xoff + wp2->sx)
-			return (wp2);
+			return (window_pane_active_set(wp, wp2));
 	}
 	return (NULL);
 }
@@ -1157,7 +1225,7 @@ window_pane_find_down(struct window_pane *wp)
 		if (wp2->yoff != bottom)
 			continue;
 		if (left >= wp2->xoff && left <= wp2->xoff + wp2->sx)
-			return (wp2);
+			return (window_pane_active_set(wp, wp2));
 	}
 	return (NULL);
 }
@@ -1186,7 +1254,7 @@ window_pane_find_left(struct window_pane *wp)
 		if (wp2->xoff + wp2->sx + 1 != left)
 			continue;
 		if (top >= wp2->yoff && top <= wp2->yoff + wp2->sy)
-			return (wp2);
+			return (window_pane_active_set(wp, wp2));
 	}
 	return (NULL);
 }
@@ -1215,7 +1283,7 @@ window_pane_find_right(struct window_pane *wp)
 		if (wp2->xoff != right)
 			continue;
 		if (top >= wp2->yoff && top <= wp2->yoff + wp2->sy)
-			return (wp2);
+			return (window_pane_active_set(wp, wp2));
 	}
 	return (NULL);
 }
@@ -1243,17 +1311,8 @@ winlink_clear_flags(struct winlink *wl)
 				continue;
 
 			wm->flags &= ~WINLINK_ALERTFLAGS;
+			wm->window->flags &= ~WINDOW_ALERTFLAGS;
 			server_status_session(s);
 		}
 	}
 }
-
-/* Set the grid_cell with fg/bg/attr information when window is in a mode. */
-void
-window_mode_attrs(struct grid_cell *gc, struct options *oo)
-{
-	memcpy(gc, &grid_default_cell, sizeof *gc);
-	colour_set_fg(gc, options_get_number(oo, "mode-fg"));
-	colour_set_bg(gc, options_get_number(oo, "mode-bg"));
-	gc->attr |= options_get_number(oo, "mode-attr");
-}
